import json
from typing import Iterable, List, Optional, Tuple

import torch
from sentence_transformers import (__MODEL_HUB_ORGANIZATION__,
                                   SentenceTransformer, __version__)
from sentence_transformers.evaluation import SentenceEvaluator
from sentence_transformers.model_card_templates import ModelCardTemplate
from sentence_transformers.util import batch_to_device, fullname
from torch import nn
from torch.utils.data import DataLoader
from tqdm import trange


def fit_sentence_transformer(
    sentence_transformer,
    train_objectives: List[Tuple[DataLoader, nn.Module]],
    evaluator: Optional[SentenceEvaluator] = None,
    epochs: int = 1,
    steps_per_epoch=None,
    scheduler: str = "WarmupLinear",
    warmup_steps: int = 10000,
    learning_rate: float = 2e-5,
    weight_decay: float = 0.01,
    max_grad_norm: float = 1,
) -> Iterable[SentenceTransformer]:
    """Custom training loop so that we can actually save stuff
    to disk and hub after each epoch.
    The function yields the current state of the model after each epoch,
    this allows top-level users to do anything with it.
    """
    optimizer_class = torch.optim.AdamW
    info_loss_functions = []
    for dataloader, loss in train_objectives:
        info_loss_functions.extend(
            ModelCardTemplate.get_train_objective_info(dataloader, loss)
        )
    info_loss_functions = "\n\n".join([text for text in info_loss_functions])

    info_fit_parameters = json.dumps(
        {
            "evaluator": fullname(evaluator),
            "epochs": epochs,
            "steps_per_epoch": steps_per_epoch,
            "scheduler": scheduler,
            "warmup_steps": warmup_steps,
            "optimizer_class": str(optimizer_class),
            "optimizer_params": {"lr": learning_rate},
            "weight_decay": weight_decay,
            "max_grad_norm": max_grad_norm,
        },
        indent=4,
        sort_keys=True,
    )
    sentence_transformer._model_card_text = None
    sentence_transformer._model_card_vars[
        "{TRAINING_SECTION}"
    ] = ModelCardTemplate.__TRAINING_SECTION__.replace(
        "{LOSS_FUNCTIONS}", info_loss_functions
    ).replace(
        "{FIT_PARAMETERS}", info_fit_parameters
    )
    sentence_transformer.to(sentence_transformer.device)
    dataloaders = [dataloader for dataloader, _ in train_objectives]
    # Use smart batching
    for dataloader in dataloaders:
        dataloader.collate_fn = sentence_transformer.smart_batching_collate
    loss_models = [loss for _, loss in train_objectives]
    for loss_model in loss_models:
        loss_model.to(sentence_transformer.device)
    sentence_transformer.best_score = -9999999
    if steps_per_epoch is None or steps_per_epoch == 0:
        steps_per_epoch = min([len(dataloader) for dataloader in dataloaders])
    num_train_steps = int(steps_per_epoch * epochs)
    # Prepare optimizers
    optimizers = []
    schedulers = []
    for loss_model in loss_models:
        param_optimizer = list(loss_model.named_parameters())
        no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p
                    for n, p in param_optimizer
                    if not any(nd in n for nd in no_decay)
                ],
                "weight_decay": weight_decay,
            },
            {
                "params": [
                    p
                    for n, p in param_optimizer
                    if any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.0,
            },
        ]
        optimizer = optimizer_class(
            optimizer_grouped_parameters, lr=learning_rate
        )
        scheduler_obj = sentence_transformer._get_scheduler(
            optimizer,
            scheduler=scheduler,
            warmup_steps=warmup_steps,
            t_total=num_train_steps,
        )
        optimizers.append(optimizer)
        schedulers.append(scheduler_obj)
    global_step = 0
    data_iterators = [iter(dataloader) for dataloader in dataloaders]
    num_train_objectives = len(train_objectives)
    skip_scheduler = False
    for epoch in trange(epochs, desc="Epoch"):
        training_steps = 0
        for loss_model in loss_models:
            loss_model.zero_grad()
            loss_model.train()
        for _ in trange(
            steps_per_epoch,
            desc="Iteration",
            smoothing=0.05,
        ):
            for train_idx in range(num_train_objectives):
                loss_model = loss_models[train_idx]
                optimizer = optimizers[train_idx]
                scheduler = schedulers[train_idx]
                data_iterator = data_iterators[train_idx]
                try:
                    data = next(data_iterator)
                except StopIteration:
                    data_iterator = iter(dataloaders[train_idx])
                    data_iterators[train_idx] = data_iterator
                    data = next(data_iterator)
                features, labels = data
                labels = labels.to(sentence_transformer.device)
                features = [
                    batch_to_device(batch, sentence_transformer.device)
                    for batch in features
                ]
                loss_value = loss_model(features, labels)
                loss_value.backward()
                torch.nn.utils.clip_grad_norm_(
                    loss_model.parameters(), max_grad_norm
                )
                optimizer.step()
                optimizer.zero_grad()
                if not skip_scheduler:
                    scheduler.step()
            training_steps += 1
            global_step += 1
        sentence_transformer._eval_during_training(
            evaluator, None, False, epoch, -1, None
        )
        yield sentence_transformer



import warnings
from typing import List, Union

from datasets import Dataset, DatasetDict
from sentence_transformers import InputExample, SentenceTransformer, losses
from sentence_transformers.evaluation import LabelAccuracyEvaluator
from torch.utils.data import DataLoader

from dfm_sentence_trf.tasks.task import Task


class Softmax(Task):
    def __init__(
        self,
        dataset: Union[Dataset, DatasetDict],
        sentence1: str,
        sentence2: str,
        label: str,
    ):
        self.dataset = dataset
        self.sentence1 = sentence1
        self.sentence2 = sentence2
        self.label = label
        if isinstance(self.dataset, Dataset):
            self.train_ds = self.dataset
            self.test_ds = None
        else:
            self.train_ds = self.dataset["train"]
            self.test_ds = self.dataset["test"]

    @property
    def n_labels(self) -> int:
        labels = self.train_ds[self.label]
        return len(set(labels))

    @property
    def examples(self) -> List[InputExample]:
        examples = []
        for entry in self.train_ds:
            example = InputExample(
                texts=[entry[self.sentence1], entry[self.sentence2]],
                label=entry[self.label],
            )
            examples.append(example)
        return examples

    @property
    def test_examples(self) -> List[InputExample]:
        if self.test_ds is None:
            return []
        shuffled = self.test_ds.shuffle()[:1000]
        examples = []
        for entry in shuffled:
            example = InputExample(
                texts=[entry[self.sentence1], entry[self.sentence2]],
                label=entry[self.label],
            )
            examples.append(example)
        return examples

    @property
    def loss(self):
        def _loss(model: SentenceTransformer):
            return losses.SoftmaxLoss(
                model=model,
                sentence_embedding_dimension=model.get_sentence_embedding_dimension(),
                num_labels=self.n_labels,
            )

        return _loss

    def evaluate(self, model: SentenceTransformer) -> float:
        if self.test_ds is None:
            warnings.warn("No test data in task, returning 0 on evaluation.")
            return 0
        model_with_loss = self.loss(model)
        test_loader = DataLoader(
            self.test_examples, shuffle=True, batch_size=16
        )
        evaluator = LabelAccuracyEvaluator(test_loader)
        return evaluator(model_with_loss)

    def __str__(self):
        # We have no reasonable way of telling if the labels mean the same,
        # so all tasks should get a different representation.
        return f"Softmax(scale={id(self)})"



from typing import Union

import catalogue
from confection import registry
from datasets import Dataset, DatasetDict

from dfm_sentence_trf.tasks.cosine_similarity import CosineSimilarity
from dfm_sentence_trf.tasks.multiplenegativeranking import (
    MultipleNegativesRanking,
)
from dfm_sentence_trf.tasks.softmax import Softmax
from dfm_sentence_trf.tasks.task import Task, to_objectives

registry.tasks = catalogue.create("confection", "tasks", entry_points=False)

__all__ = [
    "MultipleNegativesRanking",
    "Task",
    "to_objectives",
    "CosineSimilarity",
    "Softmax",
]


@registry.tasks.register("multiple_negatives_ranking")
def make_multiple_negatives_ranking(
    dataset: Union[Dataset, DatasetDict],
    sentence1: str,
    sentence2: str,
    scale: float = 20.0,
) -> MultipleNegativesRanking:
    return MultipleNegativesRanking(dataset, sentence1, sentence2, scale)


@registry.tasks.register("cosine_similarity")
def make_cosine_similarity(
    dataset: Union[Dataset, DatasetDict],
    sentence1: str,
    sentence2: str,
    similarity: str,
) -> CosineSimilarity:
    return CosineSimilarity(dataset, sentence1, sentence2, similarity)


@registry.tasks.register("softmax")
def make_softmax(
    dataset: Union[Dataset, DatasetDict],
    sentence1: str,
    sentence2: str,
    label: str,
) -> Softmax:
    return Softmax(dataset, sentence1, sentence2, label)



from confection import Config

default_config = Config(
    dict(
        training=dict(
            epochs=5,
            warmup_steps=100,
            batch_size=30,
            max_seq_length=128,
            steps_per_epoch=None,
            checkpoint_repo=None,
            wandb_project=None,
        ),
        model=dict(
            device="cpu",
            max_seq_length=128,
        ),
    ),
)



import logging
from typing import Dict

from sentence_transformers.evaluation import SentenceEvaluator

from dfm_sentence_trf.tasks.task import Task

logger = logging.getLogger(__name__)


class TaskListEvaluator(SentenceEvaluator):
    """Evaluator that evaluates the model on all test sets
    from all tasks and returns the sum of their scores."""

    def __init__(self, tasks: Dict[str, Task], log_to_wandb: bool = False):
        self.tasks = tasks
        self.log_to_wandb = log_to_wandb

    def __call__(
        self, model, output_path=None, epoch: int = -1, steps: int = -1
    ) -> float:
        scores = dict()
        for task_name, task in self.tasks.items():
            score = task.evaluate(model)
            scores[task_name] = score
        sum_score = sum(scores.values())
        scores["Summed Score"] = sum_score
        if self.log_to_wandb:
            import wandb

            wandb.log(scores)
        else:
            logger.info("Evaluation results: " + str(scores))
        return sum_score



import logging
import os
import shutil
import stat
import tempfile

from huggingface_hub import HfApi, HfFolder, Repository
from sentence_transformers import SentenceTransformer

logger = logging.getLogger(__name__)


def save_to_hub(
    model: SentenceTransformer,
    repo_name: str,
    commit_message: str = "Add new SentenceTransformer model.",
    exist_ok: bool = False,
    replace_model_card: bool = False,
):
    """Had to write this function because of a bug in sentence_transformers
    that has not been fixed in any release.
    """
    token = HfFolder.get_token()
    if token is None:
        raise ValueError(
            "You must login to the Hugging Face hub on this computer by typing `transformers-cli login`."
        )

    repo_url = HfApi().create_repo(
        repo_id=repo_name,
        token=token,
        repo_type=None,
        exist_ok=exist_ok,
    )
    _, org, name = repo_url.rsplit("/", 2)

    with tempfile.TemporaryDirectory() as tmp_dir:
        # First create the repo (and clone its content if it's nonempty).
        repo = Repository(tmp_dir, clone_from=repo_url)

        create_model_card = replace_model_card or not os.path.exists(
            os.path.join(tmp_dir, "README.md")
        )
        model.save(
            tmp_dir,
            model_name=f"{org}/{name}",
            create_model_card=create_model_card,
        )

        # Find files larger 5M and track with git-lfs
        large_files = []
        for root, dirs, files in os.walk(tmp_dir):
            for filename in files:
                file_path = os.path.join(root, filename)
                rel_path = os.path.relpath(file_path, tmp_dir)

                if os.path.getsize(file_path) > (5 * 1024 * 1024):
                    large_files.append(rel_path)

        if len(large_files) > 0:
            logger.info(
                "Track files with git lfs: {}".format(", ".join(large_files))
            )
            repo.lfs_track(large_files)

        logger.info("Push model to the hub. This might take a while")
        push_return = repo.push_to_hub(commit_message=commit_message)

        def on_rm_error(func, path, exc_info):
            # path contains the path of the file that couldn't be removed
            # let's just assume that it's read-only and unlink it.
            try:
                os.chmod(path, stat.S_IWRITE)
                os.unlink(path)
            except:
                pass

        # Remove .git folder. On Windows, the .git folder might be read-only and cannot be deleted
        # Hence, try to set write permissions on error
        try:
            for f in os.listdir(tmp_dir):
                shutil.rmtree(os.path.join(tmp_dir, f), onerror=on_rm_error)
        except Exception as e:
            logger.warning(
                "Error when deleting temp folder: {}".format(str(e))
            )
            pass

    return push_return



import logging
import warnings
from pathlib import Path
from typing import Any, Dict, Optional, Union

import catalogue
from confection import Config, registry
from datasets import Dataset, DatasetDict, load_dataset
from radicli import Arg, Radicli
from sentence_transformers import SentenceTransformer, models

from dfm_sentence_trf.config import default_config
from dfm_sentence_trf.evaluation.task_evaluator import TaskListEvaluator
from dfm_sentence_trf.hub import save_to_hub
from dfm_sentence_trf.tasks import to_objectives
from dfm_sentence_trf.training.train import fit_sentence_transformer

logger = logging.getLogger(__name__)
cli = Radicli()

registry.loaders = catalogue.create(
    "confection", "loaders", entry_points=False
)


@registry.loaders.register("load_dataset")
def loaders_load_dataset(
    path: str,
    name: Optional[str] = None,
    additional_arguments: Optional[Dict[str, Any]] = None,
) -> Union[Dataset, DatasetDict]:
    if additional_arguments is None:
        additional_arguments = dict()
    return load_dataset(path, name=name, **additional_arguments)


@cli.command(
    "finetune",
    config_path=Arg(help="Config file containing information about training."),
    output_folder=Arg(
        "--output-folder", "-o", help="Folder to save the finalized model."
    ),
    cache_folder=Arg(
        "--cache-folder",
        "-c",
        help="Folder to cache models into while training.",
    ),
)
def finetune(
    config_path: str,
    output_folder: str = "./model",
    cache_folder: str = "./model_cache",
):
    raw_config = Config().from_disk(config_path)
    raw_config = default_config.merge(raw_config)
    wandb_project = raw_config["training"].get("wandb_project")
    if wandb_project is not None:
        import wandb

        wandb.init(project=wandb_project, config=dict(raw_config))

    cfg = registry.resolve(raw_config)
    sent_trf_kwargs = dict()
    sent_trf_kwargs["device"] = cfg["model"]["device"]
    sent_trf_kwargs["cache_folder"] = cache_folder

    logger.info("Initialize SentenceTransformer model")
    embedding = models.Transformer(
        cfg["model"]["base_model"],
        max_seq_length=cfg["model"]["max_seq_length"],
    )
    pooling = models.Pooling(
        word_embedding_dimension=embedding.get_word_embedding_dimension(),
    )
    model = SentenceTransformer(
        modules=[embedding, pooling], **sent_trf_kwargs
    )

    epochs = cfg["training"]["epochs"]
    warmup_steps = cfg["training"]["warmup_steps"]
    batch_size = cfg["training"]["batch_size"]
    steps_per_epoch = cfg["training"]["steps_per_epoch"]
    checkpoint_repo = cfg["training"]["checkpoint_repo"]

    tasks = list(cfg["tasks"].values())
    evaluator = TaskListEvaluator(
        dict(cfg["tasks"]), log_to_wandb=(wandb_project is not None)
    )
    logger.info("Convert tasks to objectives")
    objectives = to_objectives(tasks, model, batch_size)
    logger.info("Starting Model Training")
    # Getting an iterable of model checkpoints after each epoch
    model_checkpoints = fit_sentence_transformer(
        model,
        objectives,
        epochs=epochs,
        warmup_steps=warmup_steps,
        evaluator=evaluator,
        steps_per_epoch=steps_per_epoch,
    )
    checkpoint_path = Path(cache_folder)
    checkpoint_path.mkdir(exist_ok=True)
    for i_epoch, checkpoint in enumerate(model_checkpoints):
        # We will save the model and push it to hub
        # if the user specifies a checkpoint repo.
        checkpoint.save(str(checkpoint_path))
        if checkpoint_repo is not None:
            try:
                save_to_hub(
                    checkpoint,
                    checkpoint_repo,
                    commit_message=f"Saved model checkpoint after epoch {i_epoch}",
                    exist_ok=True,
                    replace_model_card=True,
                )
            except ValueError as error_message:
                warnings.warn(
                    "Couldn't push model checkpoint."
                    "Did you forget to log into huggingface hub?\n"
                    f"{error_message}"
                )

    output_path = Path(output_folder)
    output_path.mkdir(exist_ok=True)
    model.save(output_folder)
    raw_config.to_disk(output_path.joinpath("config.cfg"))
    if wandb_project is not None:
        import wandb

        wandb.finish()


@cli.command(
    "push_to_hub",
    config_path=Arg(help="Config file containing information about training."),
    model_path=Arg(
        "--model_path",
        help="Path to the trained model to be pushed to the Hub.",
    ),
    commit_message=Arg(
        "--commit_message",
        "-m",
        help="Message to commit while pushing.",
    ),
    exist_ok=Arg(
        "--exist_ok",
        help="Indicates whether the model should be"
        "allowed to be pushed in an existent repo.",
    ),
    replace_model_card=Arg(
        "--replace_model_card",
        help="Indicates whether the README should be"
        "replaced with a newly generated model card.",
    ),
)
def push_to_hub(
    config_path: str,
    model_path: str,
    commit_message: str = "Add new SentenceTransformer model.",
    exist_ok: bool = False,
    replace_model_card: bool = False,
):
    raw_config = Config().from_disk(config_path)
    cfg = registry.resolve(raw_config)
    repo_name = cfg["model"]["name"]
    model = SentenceTransformer(model_path)
    save_to_hub(
        model,
        repo_name,
        commit_message=commit_message,
        exist_ok=exist_ok,
        replace_model_card=replace_model_card,
    )



from random import randrange
from typing import List


def coinflip() -> bool:
    return randrange(2) == 0


def collect_negative_ids(
    n_total: int, negative_samples: int, current_id: int
) -> List[int]:
    negative_ids = []
    for _ in range(negative_samples):
        negative_id = current_id
        while (negative_id == current_id) or (negative_id in negative_ids):
            negative_id = randrange(n_total)
        negative_ids.append(negative_id)
    return negative_ids



import warnings
from typing import List, Union

from datasets import Dataset, DatasetDict
from sentence_transformers import InputExample, SentenceTransformer, losses
from sentence_transformers.evaluation import EmbeddingSimilarityEvaluator

from dfm_sentence_trf.tasks.task import Task


class CosineSimilarity(Task):
    def __init__(
        self,
        dataset: Union[Dataset, DatasetDict],
        sentence1: str,
        sentence2: str,
        similarity: str,
    ):
        self.dataset = dataset
        self.sentence1 = sentence1
        self.sentence2 = sentence2
        self.similarity = similarity
        if isinstance(self.dataset, Dataset):
            self.train_ds = self.dataset
            self.test_ds = None
        else:
            self.train_ds = self.dataset["train"]
            self.test_ds = self.dataset["test"]

    @property
    def examples(self) -> List[InputExample]:
        examples = []
        for entry in self.train_ds:
            example = InputExample(
                texts=[entry[self.sentence1], entry[self.sentence2]],
                label=entry[self.similarity],
            )
            examples.append(example)
        return examples

    @property
    def loss(self):
        return lambda model: losses.CosineSimilarityLoss(model)

    def evaluate(self, model: SentenceTransformer) -> float:
        if self.test_ds is None:
            warnings.warn("No test data in task, returning 0 on evaluation.")
            return 0
        shuffled = self.test_ds.shuffle()[:1000]
        evaluator = EmbeddingSimilarityEvaluator(
            sentences1=shuffled[self.sentence1],
            sentences2=shuffled[self.sentence2],
            scores=shuffled[self.similarity],
        )
        return evaluator(model)

    def __str__(self):
        return "CosineSimilarity()"



from abc import abstractmethod, abstractproperty
from itertools import chain, groupby
from typing import Callable, Iterable, List, Protocol, Tuple, Union

from datasets import Dataset, DatasetDict
from sentence_transformers import InputExample, SentenceTransformer
from torch.utils.data import DataLoader


class Task(Protocol):
    dataset: Union[Dataset, DatasetDict]

    @abstractproperty
    def examples(self) -> List[InputExample]:  # type: ignore
        pass

    @abstractproperty
    def loss(self) -> Callable:  # type: ignore
        pass

    @abstractmethod
    def evaluate(self, model: SentenceTransformer) -> float:
        pass


def join_examples(tasks: Iterable[Task]) -> List[InputExample]:
    examples = (task.examples for task in tasks)
    return list(chain.from_iterable(examples))


def to_objectives(
    tasks: List[Task], model: SentenceTransformer, batch_size: int
) -> List[Tuple]:
    """Finalizes all objectives, joins all tasks that
    have the same loss function, this way the datasets from
    different objectives can be mixed in a batch."""
    objectives = []
    for _, group in groupby(tasks, key=str):
        group = list(group)
        examples = join_examples(group)
        loader = DataLoader(examples, shuffle=True, batch_size=batch_size)
        loss = group[0].loss(model)
        objectives.append((loader, loss))
    return objectives
